from __future__ import annotations

import ast
import os
import yaml

from magicgui import magic_factory
import dask.array as da
import napari
import numpy as np
import pathlib
from pathlib import Path
import tensorstore as ts
import zarr
import toolz as tz


@tz.curry
def _set_default_labels_path(widget, source_image_event):
    source_image = source_image_event.value
    if (hasattr(source_image, 'source')  # napari <0.4.8
            and source_image.source.path is not None):
        source_path = pathlib.Path(source_image.source.path)
        if source_path.suffix != '.zarr':
            labels_path = source_path.with_suffix('.zarr')
        else:
            labels_path = source_path.with_suffix('.labels.zarr')
        widget.labels_file.value = labels_path


def _on_create_labels_init(widget):
    widget.source_image.changed.connect(_set_default_labels_path(widget))


def create_ts_meta(labels_file: pathlib.Path, metadata):
    """Create bespoke metadata yaml file within zarr array."""
    fn = os.path.join(labels_file, '.naparimeta.yml')
    with open(fn, 'w') as fout:
        for key, val in metadata.items():
            if type(val) == np.ndarray:
                if np.issubdtype(val.dtype, np.floating):
                    metadata[key] = list(map(float, val))
                else:
                    metadata[key] = list(map(int, val))
        yaml.dump(metadata, fout)


def open_ts_meta(labels_file: pathlib.Path) -> dict:
    """Open bespoke metadata yaml file within zarr array, if present."""
    fn = os.path.join(labels_file, '.naparimeta.yml')
    meta = {}
    if os.path.exists(fn):
        with open(fn, 'r') as fin:
            meta = yaml.safe_load(fin)
    return meta


def open_tensorstore(labels_file: pathlib.Path, *, shape=None, chunks=None):
    if not os.path.exists(labels_file):
        zarr.open(
                str(labels_file),
                mode='w',
                shape=shape,
                dtype=np.uint32,
                chunks=chunks,
                )
    # read some of the metadata for tensorstore driver from file
    labels_temp = zarr.open(str(labels_file), mode='r')
    metadata = {
            'dtype': labels_temp.dtype.str,
            'order': labels_temp.order,
            'shape': labels_temp.shape,
            }

    dir, name = os.path.split(labels_file)
    labels_ts_spec = {
            'driver': 'zarr',
            'kvstore': {
                    'driver': 'file',
                    'path': dir,
                    },
            'path': name,
            'metadata': metadata,
            }
    data = ts.open(labels_ts_spec, create=False, open=True).result()
    return data


@magic_factory(
        labels_file={'mode': 'w'},
        widget_init=_on_create_labels_init,
        )
def create_labels(
        source_image: napari.layers.Image,
        labels_file: pathlib.Path,
        chunks='',
        ) -> napari.types.LayerDataTuple:
    """Create/load a zarr array as a labels layer based on image layer.

    Parameters
    ----------
    source_image : Image layer
        The image that we are segmenting.
    labels_file : pathlib.Path
        The path to the zarr file to be created.
    chunks : str, optional
        A string that can be evaluated as a tuple of ints specifying the chunk
        size for the zarr file. If empty, they will be (128, 128) along the
        last dimensions and (1) along any remaining dimensions. This argument
        has no effect if the file already exists.
    """
    if chunks:
        chunks_str = chunks
        chunks = ast.literal_eval(chunks)
        if (type(chunks) is not tuple
                or not all(isinstance(val, int) for val in chunks)):
            raise ValueError(
                'chunks should be a tuple of ints, e.g. "(1, 1, 512, 512)", '
                f'got {chunks_str}'
                )
    else:  # use default
        chunks = (1,) * (source_image.ndim - 2) + (128, 128)

    layer_data = open_tensorstore(
            labels_file,
            shape=source_image.data.shape,
            chunks=chunks,
            )
    layer_type = 'labels'
    layer_metadata = {
            'scale': source_image.scale,
            'translate': source_image.translate,
            }
    create_ts_meta(labels_file, layer_metadata)
    return layer_data, layer_metadata, layer_type


class LabelCorrector:
    def __init__(self,
                 image_file,
                 labels_file,
                 time_index,
                 scale=(1,1,4),
                 c=2,
                 t=None):
        """
        Correct labels to create a ground truth with five opperations,
        each of which correspond to the following number key.
            (1) toggel selection of points - to seed watershed to split
                labels.
            (2) watershed - recompute watershed for joined labels based on
                chosen points.
            (3) toggle pick label colour - choose the colour of the label
                with which to paint.
            (4) toggle label fill mode - for merging labels
            (5) toggel paint mode - paint new label.

        Output can be saved to a new file by using Control-s. Note that this is
        not necessary if using tensorstore, as annotations will be written to
        file. However, this is still useful for generating separate copies of
        GT data (if completing annotations of whole frames). If time_index is
        None, the save operation will make a copy of the currently displayed
        frame (dim = t) and the file will have a suffix of the form
        "t<idx>_GT.zarr".

        If the labels_file parameter is given a dict, this will be assumed
        to be JSON-like spec for tensorstore.
        - ASSUMES IMAGES ARE (c, t, z, y, x)
        - ASSUMES LABELS ARE (t, z, y, x)
        - ASSUMES GT FILES ARE (z, y, x)

        Parameters
        ----------
        image_file: str
            Path to the image data (.zarr)
        labels_file: str or dict
            str: Path to the labels data (.zarr)
            dict: tensorstore spec for opening the file
                (also .zarr)
        time_index: int or None
                None: all frames
                int: one frame
                TODO: slice: selection of frames
                    (must reindex ts when slicing or napari has issues)
        scale: tuple of int
            scale of image and labels for napari
        c: int
            the channel to read from the image
        t: int or none
            position of the time index in the labels file
            (if this exists). If ndim > 3 and None specified
            the format will be assumed to be (c*, t, z, y, x)

        Attributes
        ----------
        tensorstore: bool
            Should the lables be accessed via tensorstore?
            This is set to True when the inputted labels_file
            is a dict (JSON-esque spec for ts.open)
        gt_file: bool
            Is this a file ending in the suffix _GT?
            If so it is assumed to be the saved output of another
            annotation session
        time_index: slice, int, or None
            Gives the point/s in time to be selected
        labels: ndarray or tensorstore.TensorStore
        image: dask array
        scale: tuple of int
        """

        # switches
        # --------
        # tensorstore -> open zarr with tensorstore
        #   affects how we need to open data
        # gt_file -> already using a ground truth labels
        #   affects how we need to save and access the file
        #   i.e., indexes are thought to apply only to the image
        #   the ground truth is assumed to be the indexed image
        self.tensorstore = isinstance(labels_file, dict)
        self.gt_file = None # is reassigned to a bool within _get_path_info

        # Read/Write Info
        # ---------------
        self.labels_file = labels_file
        self.time_index = time_index
        self._save_path = self._get_path_info()

        # Lazy Data
        # ---------
        if time_index is None:
            self.time_index = slice(None) # ensure that the following two lines
                                          # work
        # Note: we assume a 5D array saved in ome-zarr order: tczyx
        self.image = da.from_zarr(image_file)[self.time_index, c]
        self.labels = self._open_labels()

        # Vis Info
        # --------
        self.viewer = None
        self.scale = scale
        self.ndim = len(self.image.shape)
        if self.ndim > 3 and t == None:
            t = -4
        self.t = t


    # Init helpers
    # ------------
    def _get_path_info(self):
        labels_file = self.labels_file
        # is the file storing a previously annotated and saved ground truth?
        if isinstance(labels_file, str):
            labels_path = labels_file
        elif self.tensorstore:
            # get the path str from spec
            labels_path = os.path.join(labels_file['kvstore']['path'],
                                        labels_file['path'])
        else:
            m = f'labels_file parameter must be dict or list not {type(labels_file)}'
            raise ValueError(m)
        self.gt_file = labels_path.endswith('_GT.zarr')
        save_path = self._get_save_path(labels_path)
        return save_path


    # Init helpers
    # ------------
    def _get_save_path(self, labels_path):
        if self.gt_file:
            # if working on a GT file save to the same name
            save_path = labels_path
        else:
            # otherwise use the name of the file sans extension
            # the self.save_path property will use the index on t
            # to generate the rest of the name.
            data_path = Path(labels_path)
            save_path = os.path.join(data_path.parents[0],
                                     data_path.stem)
        return save_path


    def _open_labels(self):
        labels_file = self.labels_file
        if self.tensorstore:
            # labels file should be the spec dict for tensorstore
            if not self.gt_file:
                # we need to apply the slice and so need to construct
                # the correct tuple of int / slices
                labels = labels[self.time_index]
        else:
            labels = zarr.open(labels_file, mode='r+')
            if not self.gt_file:
                labels = labels[self.time_index]
        return labels


    # CALL
    # ----
    def __call__(self):
        with napari.gui_qt():
            self.viewer =  napari.Viewer()
            self.viewer .add_image(
                                   self.image,
                                   name='Image',
                                   scale=self.scale
                                   )
            self.viewer .add_labels(
                                    self.labels,
                                    name='Labels',
                                    scale=self.scale
                                    )
            self.viewer .add_points(
                                    np.empty(
                                             (0, len(self.labels.shape)),
                                             dtype=float
                                             ),
                                    scale=self.scale,
                                    size=2)
            self.viewer .bind_key('1', self._points)
            self.viewer .bind_key('2', self._watershed)
            self.viewer .bind_key('3', self._select_colour)
            self.viewer .bind_key('4', self._fill)
            self.viewer .bind_key('5', self._paint)
            self.viewer .bind_key('Shift-s', self._save)


    @property
    def save_path(self):
        if self.gt_file:
            # no need to find the save path suffix, it is ex
            return self._save_path
        else:
            if isinstance(self.time_index, int):
                time = self.time_index
            elif self.ndim > 3:
                time = self.viewer.dims.current_step[self.t]
            else:
                time = 'unknown'
            suffix = f"_t{time}_GT.zarr"
            return self._save_path + suffix


    # Call helpers
    # ------------
    def _points(self, viewer):
        """
        Switch to points layer to split a label
        """
        if viewer.layers['Points'].mode != 'add':
            viewer.layers['Points'].mode = 'add'
        else:
            viewer.layers['Points'].mode = 'pan_zoom'




    def _select_colour(self, viewer):
        """
        Select colour for painting
        """
        if viewer.layers['Labels'].mode != 'pick':
            viewer.layers['Labels'].mode = 'pick'
        else:
            viewer.layers['Labels'].mode = 'pan_zoom'


    def _fill(self, viewer):
        """
        Switch napari labels layer to fill mode
        """
        if viewer.layers['Labels'].mode != 'fill':
            viewer.layers['Labels'].mode = 'fill'
        else:
            viewer.layers['Labels'].mode = 'pan_zoom'


    def _paint(self, viewer):
        """
        Switch napari labels layer to paint mode
        """
        if viewer.layers['Labels'].mode != 'paint':
            viewer.layers['Labels'].mode = 'paint'
        else:
            viewer.layers['Labels'].mode = 'pan_zoom'


    def _save(self, viewer):
        """
        Save the annotated time_index as a zarr file
        """
        array = viewer.layers['Labels'].data
        if len(array.shape) > 3: # save the current frame
            time = self.viewer.dims.current_step[self.t]
            idx = [slice(None)] * self.ndim
            idx[self.t] = time
            idx = tuple(idx)
            array = array[idx]
        zarr.save_array(self.save_path, np.array(array))
        print("Labels saved at:")
        print(self.save_path)


# Split Objects
# -------------


import os
import tempfile
import napari
import numpy as np
from numpy.testing import assert_array_equal
from zarpaint import create_labels, DimsSorter
from zarpaint.plugin import zarr_tensorstore


def test_create_labels():
    image = napari.layers.Image(
        np.random.random((10, 512, 512)),
        name='img',
        scale=[4, 1, 1],
        translate=[1000, -1000, -2000],
        )
    create_labels_ = create_labels()  # instance from magic_factory
    with tempfile.TemporaryDirectory() as tmpdir:
        pth = os.path.join(tmpdir, 'labels.zarr')
        arr, meta, layer_type = create_labels_(image, pth, str((1, 128, 128)))
        # note that arr.shape is an array if arr is tensorstore.Array
        assert tuple(arr.shape) == image.data.shape
        for key, val in meta.items():
            assert_array_equal(getattr(image, key), val)
        assert layer_type == 'labels'
        assert os.path.exists(pth)
        assert len(os.listdir(pth)) == 2  # .zarray, .naparimeta.yml
        arr[4, :256, :256] = 1  # touch 4 chunks
        assert len(os.listdir(pth)) == 6


def test_dims_sorter(make_napari_viewer):
    viewer = make_napari_viewer(strict_qt=False)
    ndim = 5
    viewer.add_points(np.random.random((10, ndim)) * 512)
    sorter = DimsSorter(viewer)
    new_order = (1, 0, 2, 3, 4)
    viewer.dims.order = new_order
    assert [sorter.axes_list[i].axis for i in range(ndim)] == list(new_order)
    assert sorter.axes_list[0].dims is viewer.dims
    sorter.axes_list.move(4, 3)
    assert viewer.dims.order == (1, 0, 2, 4, 3)


def test_open_tensorstore():
    image = napari.layers.Image(
        np.random.random((10, 512, 512)),
        name='img',
        scale=[4, 1, 1],
        translate=[1000, -1000, -2000],
        )
    create_labels_ = create_labels()  # instance from magic_factory
    with tempfile.TemporaryDirectory() as tmpdir:
        pth = os.path.join(tmpdir, 'labels.zarr')
        _ = create_labels_(image, pth, str((1, 128, 128)))
        # note that arr.shape is an array if arr is tensorstore.Array
        reader = zarr_tensorstore(pth)
        arr, meta, layer_type = reader(pth)[0]  # 1st element of layer list
        assert tuple(arr.shape) == image.data.shape
        for key, val in meta.items():
            assert_array_equal(getattr(image, key), val)
        assert layer_type == 'labels'
        assert os.path.exists(pth)
        assert len(os.listdir(pth)) == 2  # .zarray, .naparimeta.yml
        arr[4, :256, :256] = 2  # touch 4 chunks
        assert len(os.listdir(pth)) == 6



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"


from ._zarpaint import create_labels, open_tensorstore
from ._dims_chooser import DimsSorter, set_axis_labels
from ._watershed import watershed_split

__all__ = [
    'create_labels',
    'open_tensorstore',
    'DimsSorter',
    'set_axis_labels',
    'watershed_split',
]



from __future__ import annotations

import os
import pathlib
from napari_plugin_engine import napari_hook_implementation
from ._zarpaint import create_labels, open_tensorstore, open_ts_meta
from ._dims_chooser import DimsSorter, set_axis_labels
from ._watershed import watershed_split


@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    return create_labels


@napari_hook_implementation(
    specname='napari_experimental_provide_dock_widget'
)
def dims_sorter():
    return DimsSorter


@napari_hook_implementation(
    specname='napari_experimental_provide_dock_widget'
)
def axis_labels():
    return set_axis_labels


@napari_hook_implementation(
    specname='napari_experimental_provide_dock_widget'
)
def watershed():
    return watershed_split


@napari_hook_implementation(
    specname='napari_get_reader'
)
def zarr_tensorstore(path: str | pathlib.Path):
    if (str(path).endswith('.zarr')
            and os.path.isdir(path) and '.zarray' in os.listdir(path)):
        return lambda path: [(open_tensorstore(path), open_ts_meta(path), 'labels')]



from __future__ import annotations

from magicgui import magic_factory
import numpy as np
import napari
from qtpy.QtWidgets import QStyledItemDelegate
from qtpy.QtGui import QBrush
from napari.utils.events import SelectableEventedList
from qtpy.QtWidgets import QWidget, QHBoxLayout


class AxisModel:
    """View of an axis within a dims model keeping track of axis names."""

    def __init__(self, dims: napari.components.Dims, axis: int):
        self.dims = dims
        self.axis = axis

    def __hash__(self):
        return id(self)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return self.dims.axis_labels[self.axis]

    def __eq__(self, other: int | str):
        if isinstance(other, int):
            return self.axis == other
        else:
            return repr(self) == other


# this class is currently unused — needed to paint visible dims differently
class DimsDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        if index.row() > 3:
            brush = QBrush()
            brush.setColor('red')
            option.backgroundBrush = brush
        super().paint(painter, option, index)


def set_dims_order(dims, order):
    if type(order[0]) == AxisModel:
        order = [a.axis for a in order]
    dims.order = order


def _array_in_range(arr, low, high):
    return (arr >= low) & (arr < high)


def move_indices(axes_list, order):
    with axes_list.events.blocker_all():
        axes = [a.axis for a in axes_list]
        if tuple(axes_list) == tuple(order):
            return
        ax_to_existing_position = {a: ix for ix, a in enumerate(axes)}
        move_list = np.asarray([
            (ax_to_existing_position[order[i]], i)
            for i in range(len(order))
        ])
        for src, dst in move_list:
            axes_list.move(src, dst)
            move_list[_array_in_range(move_list[:, 0], dst, src)] += 1
        # remove the elements from the back if order has changed length
        while len(axes_list) > len(order):
            axes_list.pop()


class DimsSorter(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', parent=None):
        super().__init__(parent=parent)
        from napari._qt.containers import QtListView
        dims = napari_viewer.dims
        root = SelectableEventedList(
            [AxisModel(dims, i) for i in range(dims.ndim)]
        )
        root.events.reordered.connect(
            lambda event, dims=dims: set_dims_order(dims, event.value)
        )
        dims.events.order.connect(
            lambda event, axes_list=root: move_indices(axes_list, event.value)
        )
        view = QtListView(root)
        self.axes_list = root
        self.setLayout(QHBoxLayout())
        self.layout().addWidget(view)


@magic_factory(
    call_button='set axis labels',
    viewer={'visible': False},
)
def set_axis_labels(viewer: napari.Viewer, axes=''):
    if type(axes) == str and len(axes) == viewer.dims.ndim:
        viewer.dims.axis_labels = list(axes)


if __name__ == '__main__':
    from napari._qt.containers import QtListView
    from napari.components import Dims

    dims = Dims(ndim=5, ndisplay=2, last_used=0, axis_labels=list('tzcyx'))
    # create a python model
    root = SelectableEventedList([AxisModel(dims, i) for i in range(5)])
    root.events.reordered.connect(
        lambda event, dims=dims: set_dims_order(dims, event.value)
    )
    # note: we don't yet handle expanding the order
    dims.events.order.connect(
        lambda event, axes_list=root: move_indices(axes_list, event.value)
    )

    # create Qt views onto the python models
    view = QtListView(root)

    w = QWidget()
    w.setLayout(QHBoxLayout())
    w.layout().addWidget(view)
    w.show()

    dims.order = (4, 3, 1, 0, 2)  # xyztc

    napari.run()



import numpy as np
from scipy import ndimage as ndi
from magicgui import magic_factory
from skimage.morphology import octahedron
from skimage.segmentation import watershed


def slice_points(points_layer, dims, ndim):
    tf = points_layer._transforms['data2world'].inverse
    steps = tf.scale[:-ndim] * np.array(dims.range)[:-ndim, 2]
    pt = tf(np.array(dims.point))[:-ndim]
    data = points_layer.data
    sel = np.ones(data.shape[0], dtype=bool)
    for i, (p, st) in enumerate(zip(pt, steps)):
        sel &= p - st/2 <= data[:, i]
        sel &= data[:, i] < p + st/2
    return data[sel]


@magic_factory(
    call_button='Split',
    viewer={'visible': False},
    ndim={'min': 2, 'max': 3},
)
def watershed_split(
        viewer: 'napari.viewer.Viewer',
        labels: 'napari.layers.Labels',
        points: 'napari.layers.Points',
        ndim: int = 3,
        ):
    """Execute watershed to split labels based on provided points.

    Only labels containing points will be modified. The operation will be
    performed in-place on the input Labels layer.

    Parameters
    ----------
    viewer : napari.Viewer
        The current viewer displaying the data.
    labels : napari Labels layer
        The layer containing the segmentation. This will be used both for
        input and output.
    points : napari Points layer
        The points marking the labels to be split.
    ndim : int in {2, 3}
        The number of dimensions for the watershed operation.
    """
    coord = viewer.dims.current_step
    slice_idx = coord[:-ndim]
    # find the labels corresponding to the current points in the points layer
    labels_sliced = np.asarray(labels.data[slice_idx])
    points_sliced = slice_points(points, viewer.dims, ndim)
    points_data_to_world = points._transforms['data2world']
    labels_world_to_data = labels._transforms['data2world'].inverse
    points_transformed = labels_world_to_data(
            points_data_to_world(points_sliced)
            ).astype(int)[:, -ndim:]
    image = np.ones_like(labels_sliced, dtype=np.uint8)
    labels_split = _watershed_split(
            image,
            labels_sliced,
            points_transformed,
            compactness=200,
            connectivity_octahedron=7,
            )
    labels.data[slice_idx] = labels_split
    points.data = np.empty((0, viewer.dims.ndim), dtype=float)
    labels.refresh()


def _watershed_split(
        image,
        labels,
        points,
        compactness=200,
        connectivity_octahedron=7
        ):
    """
    Split labels with using points as markers for watershed
    """
    connectivity = octahedron(connectivity_octahedron)
    points = np.round(points).astype(int)
    coords = tuple([points[:, i] for i in range(points.shape[1])])
    p_lab = labels[coords]
    p_lab = np.unique(p_lab)
    p_lab = p_lab[p_lab != 0]
    # generate a mask corresponding to the labels that need to be split
    mask = np.zeros(labels.shape, dtype=bool)
    for lab in p_lab:
        where = labels == lab
        mask = mask + where
    # split the labels using the points (in the masked image)
    markers_bool = np.zeros(labels.shape, dtype=bool)
    markers_bool[coords] = True
    markers, _ = ndi.label(markers_bool, output=labels.dtype)
    new_labels = watershed(
            image,
            markers=markers,
            mask=mask,
            compactness=compactness,
            connectivity=connectivity,
            )
    labels[mask] = new_labels[mask] + labels.max()
    return labels



import argparse
import ast
import sys

from ._zarpaint import correct_labels


parser = argparse.ArgumentParser(
        'zarpaint',
        description='Paint segmentations directly to '
                    'on-disk/remote zarr arrays',
        )
parser.add_argument('image', help='The input image file.')
parser.add_argument('labels', help='The labels file (read/write).')
parser.add_argument('--frame', type=int, help='Load only this frame.')
parser.add_argument(
        '-s', '--scale',
        type=ast.literal_eval,
        default=(4, 1, 1),
        help='Scale factors.',
        )
parser.add_argument(
        '-c', '--channel',
        type=int,
        default=2,
        help='Which channel to load.',
        )


def main():
    args = parser.parse_args(sys.argv[1:])
    correct_labels(
            args.image, args.labels,
            time_index=args.frame,
            scale=args.scale,
            c=args.channel,
            )
